
from csp import *
import time

#______________________________________________________________________________
# Kakuro Problems

def kakuro_constraint(A, a, B, b) :
    """Constraint is satisfied (true) if A, B are different and if the numbers
    in this group selected so far are compatible with the constrain of the total sum."""

    kakCsp.consistencyChecks += 1
    
    #If A and B have the same value the constraint is not satisfied
    if a == b :
        return False

    #Get numbers of row and column for A and B
    AX = A.split(",")[0]
    BX = B.split(",")[0]
    AY = A.split(",")[1]        
    BY = B.split(",")[1]

    #If A and B are horizontal neighbors
    if AX == BX :
        return check_total_sum(A, a, B, b, kakCsp.varDict[A].horN, kakCsp.varDict[A].horSum)
            
    #If A and B are vertical neighbors
    else :
        return check_total_sum(A, a, B, b, kakCsp.varDict[A].verN, kakCsp.varDict[A].verSum)
        

def check_total_sum(A, a, B, b, neighbors, totalSum) :
    """For the compatibility check of the total sum, a list of all possible values
    is initialized and every time a number is used it is removed from the list. In
    this way we will have a list of remaining values, if not all the variables have
    a value, to check if any of these can satisfy the constrain of the total sum.
    Return True if the constrain is satisfied."""
    
    #Initialize the possible values list and remove a and b
    dom = set(range(1, 10))
    dom.discard(a)
    dom.discard(b)
    
    curSum = a + b;
    noneSum = 0

    if curSum > totalSum :
        return False

    #Iterate through all the horizontal neighbors
    for N in neighbors :
        #If N isn't B and N has a value
        if N != B and kakCsp.varDict[N].currentValue != None :
            curSum += kakCsp.varDict[N].currentValue
            if curSum > totalSum :
                return False
            dom.discard(kakCsp.varDict[N].currentValue)
        #If N isn't B and N has no value
        elif N != B and kakCsp.varDict[N].currentValue == None :
            noneSum += 1

    #if all the neibors have a value check but the sum is not right
    if noneSum == 0 and curSum != totalSum :
        return False;        

    #Sort the remaining possible values
    dom = sorted(dom)

    #Check if by adding the smallest possible values the sum is exceeded
    minSum = curSum
    for i in range(noneSum) :
        minSum += dom[i]
    if minSum > totalSum :
        return False

    #Check if by adding the biggest possible values the sum is not reached
    maxSum = curSum
    for i in range(noneSum) :
        maxSum += dom[len(dom)-i-1]
    if maxSum < totalSum :
        return False

    return True
    
    
def parse_setup(setup) :
    """Parses the initial setup of the kakuro puzzle. The structure of the input
    is as follows:
    (Start Row number),(Start column number)-(End Row number),(End column number):SUM;
    where the row and column numbers are starting from 1 and the SUM is the number of
    that defines the total of this line of squares. Example: 2,2-2,4:5;"""
    
    setup = setup.split(";")

    varDict = {}
    for line in setup :
        if line == "" :
            break
        
        line = line.strip()
        line = line.split(":")
        start = line[0].split("-")[0]
        end = line[0].split("-")[1]
        lineSum = line[1]

        startX = start.split(",")[0]
        endX = end.split(",")[0]
        startY = start.split(",")[1]        
        endY = end.split(",")[1]
        
        if startX == endX :
            for i in range(int(startY), int(endY) + 1) :
                key = startX + "," + str(i)

                neighbors = []
                for j in range(int(startY), int(endY) + 1) :
                    if i != j :
                        neighbors.append(startX + "," + str(j))
                
                if key not in varDict :
                    varDict[key] = KakuroSquare()

                varDict[key].horSum = int(lineSum)
                varDict[key].horN.update(neighbors)
        else :
            for i in range(int(startX), int(endX) + 1) :
                key = str(i) + "," + startY

                neighbors = []
                for j in range(int(startX), int(endX) + 1) :
                    if i != j :
                        neighbors.append(str(j) + "," + startY)
                
                if key not in varDict :
                    varDict[key] = KakuroSquare()

                varDict[key].verSum = int(lineSum)
                varDict[key].verN.update(neighbors)
                
    return varDict

##------------------------------Kakuro CSP Class------------------------------

class KakuroCSP(CSP) :
    """Class that is used to solve Kakuro csp puzzles"""
    
    def __init__(self, rows, columns, setup) :
        """Initialize data structures for a rows x columns Kakuro puzzle
        that has structure based on the setup."""

        self.consistencyChecks = 0
        self.rows = rows
        self.columns = columns
        self.varDict = parse_setup(setup)

        #For each variable unite the sets of horizontal and vertical neighbors
        neighbors = {}
        for key, value in self.varDict.iteritems() :
            neighbors[key] = list(value.verN.union(value.horN))

        CSP.__init__(self, neighbors.keys(), UniversalDict(range(1, 10)), neighbors,
               kakuro_constraint)

    def assign(self, var, val, assignment) :
        """Assign val the variable for use in the constraint check."""
        
        self.varDict[var].currentValue = val
        
        CSP.assign(self, var, val, assignment)

    def unassign(self, var, assignment) :
        """Set the value of the variable to None."""
        
        self.varDict[var].currentValue = None
        
        CSP.unassign(self, var, assignment)

    def printSolution(self, result) :
        """Prints the result on a grid like a real Kakuro and some statistics"""
        
        for i in range(self.rows) :
            curRow = []
            for j in range(self.columns) :
                key = str(i+1)+","+str(j+1)
                if key in result :
                    curRow.append(result[key])
                    del result[key]
                else :
                    curRow.append(0)
            print curRow
        print ""              

##------------------------------Kakuro Square Class------------------------------
        
class KakuroSquare() :
    """Class that stores the data for each square that is a variable"""
    
    def __init__(self) :
        self.currentValue = None
        self.verN = set()
        self.horN = set()
        self.verSum = 0
        self.horSum = 0


##------------------------------Kakuro Setup Examples------------------------------

hwExample = (7, 7, """  2,3-2,4:4; 2,3-3,3:4; 2,4-5,4:10;
                        3,3-3,4:3; 3,6-3,7:4; 3,6-4,6:3; 3,7-4,7:4;
                        4,4-4,7:10; 4,5-7,5:11;
                        5,2-5,5:11; 5,2-6,2:3; 5,3-6,3:4;
                        6,2-6,3:4; 6,5-6,6:4; 6,6-7,6:4;
                        7,5-7,6:3;""")

small = (8, 6, """  2,3-2,5:7; 2,3-8,3:28; 2,4-3,4:4; 2,5-8,5:42;
                    3,2-3,6:32; 3,2-4,2:7; 3,6-4,6:9;
                    4,2-4,3:3; 4,5-4,6:6;
                    5,3-5,5:11; 5,4-6,4:9;
                    6,3-6,5:19;
                    7,2-7,3:12; 7,2-8,2:16; 7,5-7,6:5; 7,6-8,6:11;
                    8,2-8,3:15; 8,5-8,6:16;""")

medium = (10, 8, """2,2-2,3:8; 2,2-3,2:16; 2,3-3,3:3; 2,6-2,7:6; 2,6-3,6:5; 2,7-7,7:38;
                    3,2-3,8:41; 3,4-4,4:6; 3,5-9,5:41; 3,8-4,8:7;
                    4,4-4,5:3; 4,7-4,8:4;
                    5,2-5,3:14; 5,2-6,2:15; 5,3-10,3:22; 5,5-5,7:23; 5,6-6,6:12; 
                    6,2-6,8:29; 6,4-7,4:4; 6,8-7,8:3;
                    7,3-7,5:7; 7,7-7,8:7;
                    8,2-8,3:4; 8,2-9,2:3; 8,5-8,6:10; 8,6-9,6:6;
                    9,2-9,8:41; 9,4-10,4:7; 9,7-10,7:9; 9,8-10,8:15;
                    10,3-10,4:8; 10,7-10,8:7;""")

large = (12, 10, """ 2,4-2,6:11; 2,4-3,4:6; 2,5-3,5:10; 2,6-4,6:22; 2,9-2,10:15; 2,9-8,9:30; 2,10-4,10:20;
                    3,2-3,7:38; 3,2-4,2:16; 3,3-4,3:8; 3,7-4,7:5; 3,9-3,10:13;
                    4,2-4,3:11; 4,6-4,10:15; 4,8-6,8:14;
                    5,4-5,5:7; 5,4-6,4:7; 5,5-7,5:23; 5,8-5,9:12;
                    6,2-6,6:34; 6,2-7,2:10; 6,3-12,3:28; 6,6-8,6:11; 6,8-6,9:3;
                    7,2-7,3:5; 7,5-7,7:11; 7,7-9,7:8; 7,9-7,10:6;
                    8,3-8,4:15; 8,4-10,4:21; 8,6-8,10:32; 8,8-9,8:16;
                    9,3-9,4:13; 9,7-9,8:8;
                    10,2-10,6:16; 10,2-12,2:23; 10,5-11,5:12; 10,6-12,6:6; 10,9-10,10:6; 10,9-11,9:6; 10,10-11,10:12;
                    11,2-11,3:9; 11,5-11,10:38; 11,7-12,7:9; 11,8-12,8:11;
                    12,2-12,3:16; 12,6-12,8:8;""")


selkakDict = {'1' : hwExample, '2' : small, '3' : medium, '4' : large}


#______________________________________________________________________________
# Main


while True :
    print "Please select one of the following puzzle by typing the respective number:\n"
    print "1. Homework example"
    print "2. Small sized"
    print "3. Medium sized"
    print "4. Large sized"
    print "\n0. Exit\n"

    selPuzzle = raw_input('--> ')
    if selPuzzle == '0' :
        break
    elif selPuzzle in selkakDict :
        print "Please select one of the following algorithms by typing the respective number:\n"
        print "1. BT"
        print "2. BT+MRV"
        print "3. FC"
        print "4. FC+MRV"
        print "\n0. Exit\n"

        
        sel = raw_input('--> ')
        try:
            sel = int(sel)
        except ValueError:
            print "Wrong input!! Please try again...\n"
            continue
            
        if sel == 0 :
            break
        elif sel in range(1, 5) :
            start_time = time.time()
            
            kakCsp = KakuroCSP(selkakDict[selPuzzle][0], selkakDict[selPuzzle][1], selkakDict[selPuzzle][2])

            result = None
            if sel == 1 :
                result = backtracking_search(kakCsp)
            if sel == 2 :
                result = backtracking_search(kakCsp, select_unassigned_variable=mrv)
            if sel == 3 :
                result = backtracking_search(kakCsp, inference=forward_checking)
            if sel == 4 :
                result = backtracking_search(kakCsp, select_unassigned_variable=mrv, inference=forward_checking)

            if result != None :
                kakCsp.printSolution(result)
                print "Time elapsed: {0} sec\nNumber of assignments: {1}\nNumber of consistency checks: {2}\n".format(
                        time.time() - start_time, kakCsp.nassigns, kakCsp.consistencyChecks)
            else :
                print "No solution was found. Please check your input values.\n"
        else :
            print "Wrong input!! Please try again...\n"
            continue
    else :
        print "Wrong input!! Please try again...\n"
        continue
